from typing import Any
import logging
import json

from archetypeai._base import ApiBase
from archetypeai._socket_manager import SocketManager


class SensorsApi(ApiBase):
    """Main sensor client for streaming data to the Archetype AI platform."""

    def __init__(self, api_key: str, api_endpoint: str, num_sensor_threads: int = 1) -> None:
        super().__init__(api_key, api_endpoint)
        self.num_sensor_threads = num_sensor_threads
        self.streamer = None
        self.subscribers = []
    
    def register(self, sensor_name: str, sensor_metadata: dict = {}, topic_ids: list[str] = []) -> bool:
        """Registers a sensor with the Archetype AI platform."""
        api_endpoint = self._get_endpoint(self.api_endpoint, "sensors", "register")
        data_payload = {"sensor_name": sensor_name, "sensor_metadata": sensor_metadata, "topic_ids": topic_ids}
        response = self.requests_post(api_endpoint, data_payload=json.dumps(data_payload))
        logging.info(f"Successfully registered sensor {sensor_name} stream_uid: {response['stream_uid']}")
        self.streamer = SocketManager(self.api_key, self.api_endpoint, num_worker_threads=self.num_sensor_threads)
        self.streamer._start_stream(response["stream_uid"], response["sensor_endpoint"], "sensors/streamer")
        return True
    
    def subscribe(self, sensor_name: str, topic_ids: list[str] = []) -> bool:
        """Subscribes to a sensor stream from the Archetype AI platform."""
        api_endpoint = self._get_endpoint(self.api_endpoint, "sensors", "subscribe")
        data_payload = {"sensor_name": sensor_name, "topic_ids": topic_ids}
        response = self.requests_post(api_endpoint, data_payload=json.dumps(data_payload))
        logging.info(f"Successfully subscribed to sensor {sensor_name} subscriber_uid: {response['subscriber_uid']}")
        subscriber = SocketManager(
            self.api_key, self.api_endpoint, num_worker_threads=self.num_sensor_threads, fetch_time_sec=0.1)
        subscriber._start_stream(response["subscriber_uid"], response["subscriber_endpoint"], "sensors/subscriber")
        self.subscribers.append(subscriber)
        return True

    def send(self, topic_id: str, data: Any, timestamp: float = -1.0) -> bool:
        assert self.streamer is not None, "Sensor not registered. Call register first."
        success = self.streamer.send(topic_id, data, timestamp)
        return success

    def close(self) -> bool:
        if self.streamer:
            self.streamer.close()
        for subscriber in self.subscribers:
            subscriber.close()
        return True 

    def get_stats(self):
        assert self.streamer is not None, "Sensor not registered. Call register first."
        return self.streamer.get_stats()

    def get_incoming_data_queue_size(self) -> int:
        assert self.streamer is not None, "Sensor not registered. Call register first."
        return self.streamer.incoming_data_queue.qsize()
    
    def get_incoming_message_queue_size(self) -> int:
        assert self.streamer is not None, "Sensor not registered. Call register first."
        return self.streamer.incoming_message_queue.qsize()
    
    def get_outgoing_message_queue_size(self) -> int:
        assert self.streamer is not None, "Sensor not registered. Call register first."
        return self.streamer.outgoing_message_queue.qsize()
    
    def get_max_outgoing_message_queue_size(self) -> int:
        assert self.streamer is not None, "Sensor not registered. Call register first."
        return self.streamer.max_outgoing_message_queue_size
    
    def get_outgoing_message_queue_latency(self) -> float:
        """Returns the latency of the latest outgoing message queue in seconds."""
        assert self.streamer is not None, "Sensor not registered. Call register first."
        return self.streamer.get_outgoing_message_queue_latency()

    def get_sensor_data(self) -> list[dict]:
        events = []
        for subscriber in self.subscribers:
            for sensor_name, topic_id, data in subscriber.get_data():
                events.append({"sensor_name": sensor_name, "topic_id": topic_id, "data": data})
        return events


import json

from archetypeai._base import ApiBase


class CapabilitiesApi(ApiBase):
    """Main class for handling all capability API calls."""
    
    def summarize(self, query: str, file_ids: list[str]) -> dict:
        """Runs the summarization API on the list of file IDs."""
        api_endpoint = self._get_endpoint(self.api_endpoint, "summarize")
        data_payload = {"query": query, "file_ids": file_ids}
        response_data = self.requests_post(api_endpoint, data_payload=json.dumps(data_payload))
        return response_data

    def describe(self, query: str, file_ids: list[str]) -> dict:
        """Runs the description API on the list of file IDs."""
        api_endpoint = self._get_endpoint(self.api_endpoint, "describe")
        data_payload = {"query": query, "file_ids": file_ids}
        response_data = self.requests_post(api_endpoint, data_payload=json.dumps(data_payload))
        return response_data


import logging
import json
import time
from queue import Queue
from typing import Any
import threading

from websocket import create_connection

from archetypeai._base import ApiBase

_CTRL_MSG_HEADER = "cm"
_DATA_MSG_HEADER = "dm"
_HEADER_KEY = "h"
_HEARTBEAT_DELAY_SEC = 5.0


class SocketManager(ApiBase):
    """Helper class for communicating with the Archetype AI platform via websockets."""

    def __init__(self, api_key: str, api_endpoint: str, num_worker_threads: int = 1, fetch_time_sec=2.0) -> None:
        super().__init__(api_key, api_endpoint)
        self.stream_uid = None
        self.streamer_endpoint = None
        self.num_workers = num_worker_threads
        self.fetch_time_sec = fetch_time_sec
        self.connected = False
        self.streamer_sockets = []
        self.threads = {}
        self.post_connect_timeout_sec = 1
        self.incoming_data_queue = Queue()
        self.incoming_message_queue = Queue()
        self.outgoing_message_queue = Queue()
        self.stats_queue = Queue()
        self.message_id = 0
        self._run_worker_loop = False
        self.max_outgoing_message_queue_size = 0
        self.outgoing_message_latency_total = 0.0
        self.outgoing_message_count = 0
        self.stats = {}
        self.stats["num_data_packets_sent"] = 0
        self.stats["max_outgoing_message_queue_size"] = 0
        self.stats["outgoing_message_queue_latency"] = 0
        self.stats["outgoing_message_latency"] = 0
    
    def _start_stream(self, stream_uid: str, streamer_endpoint: str, streamer_channel: str) -> bool:
        """Starts a new stream with the Archetype AI platform."""
        self.stream_uid = stream_uid
        self.streamer_endpoint = streamer_endpoint
        self.streamer_channel = streamer_channel
        self.message_id = 0
        
        self._safely_stop_streams()
        assert self._handshake()
        for worker_id, streamer_socket in enumerate(self.streamer_sockets):
            self.threads[worker_id] = threading.Thread(target=self._worker, args=(worker_id, streamer_socket))
            self.threads[worker_id].start()

        self.connected = True
        return self.connected

    def _safely_stop_streams(self):
        self._run_worker_loop = False
        for worker_id in self.threads:
            self.threads[worker_id].join()
        self.threads = {}
        self.streamer_sockets = []
        self.connected = False

    def close(self, wait_on_pending_data: bool = True):
        """Closes the connection with the server."""
        if wait_on_pending_data:
            while not self.outgoing_message_queue.empty():
                logging.info(f"outgoing data queue size: {self.outgoing_message_queue.qsize()}")
                time.sleep(0.1)
        self._safely_stop_streams()

    def send(self, topic_id: str, data: Any, timestamp: float = -1.0) -> bool:
        """Sends data to the Archetype AI platform under the given topic_id."""
        assert self.connected, "Client not connected. Make sure the stream is open!"
        timestamp = timestamp if timestamp >= 0 else time.time()
        message = {
            "topic_id": topic_id,
            "data": data,
            "timestamp": timestamp,
            "message_id": self.message_id,
            "stream_uid": self.stream_uid
        }
        self.message_id += 1
        self.outgoing_message_queue.put({_HEADER_KEY: _DATA_MSG_HEADER, **message})
        self._refresh_stats()
        return True

    def get_messages(self) -> Any:
        """Gets any pending messages sent to the client."""
        assert self.connected, "Client not connected. Make sure the stream is open!"
        while not self.incoming_message_queue.empty():
            topic_id, data = self.incoming_message_queue.get()
            yield topic_id, data

    def get_data(self) -> Any:
        """Gets any pending data sent to the client."""
        assert self.connected, "Client not connected. Make sure the stream is open!"
        while not self.incoming_data_queue.empty():
            sensor_name, topic_id, data = self.incoming_data_queue.get()
            yield sensor_name, topic_id, data

    def get_incoming_data_queue_size(self) -> int:
        return self.incoming_data_queue.qsize()
    
    def get_incoming_message_queue_size(self) -> int:
        return self.incoming_message_queue.qsize()
    
    def get_outgoing_message_queue_size(self) -> int:
        return self.outgoing_message_queue.qsize()
    
    def get_max_outgoing_message_queue_size(self) -> int:
        return self.stats["max_outgoing_message_queue_size"]
    
    def get_outgoing_message_queue_latency(self) -> float:
        """Returns the latency of the latest outgoing message queue in seconds."""
        self._refresh_stats()
        return self.stats["outgoing_message_queue_latency"]
    
    def get_outgoing_message_latency(self) -> float:
        """Returns the average latency of outgoing data packets in seconds."""
        self._refresh_stats()
        return self.stats["outgoing_message_queue_latency"]
    
    def get_stats(self) -> dict:
        """Returns the stats of a sensor stream."""
        self._refresh_stats()
        return self.stats

    def _refresh_stats(self):
        while not self.stats_queue.empty():
            stats_event = self.stats_queue.get()
            if "outgoing_message_queue_latency" in stats_event:
                self.stats["outgoing_message_queue_latency"] = stats_event["outgoing_message_queue_latency"]
            elif "max_outgoing_message_queue_size" in stats_event:
                self.stats["max_outgoing_message_queue_size"] = max(
                    stats_event["max_outgoing_message_queue_size"], self.stats["max_outgoing_message_queue_size"])

    def _worker(self, worker_id: str, streamer_socket) -> None:
        logging.debug(f"Starting worker {worker_id}")
        self._run_worker_loop = True
        try:
            self._worker_loop(worker_id, streamer_socket)
        except:
            logging.exception(f"Main loop failed, closing socket!")
            # Remove this failed worker from the thread pool.
            del self.threads[worker_id]
            del streamer_socket
            # If this worker has stopped then make sure all workers stop.
            self._safely_stop_streams()
    
    def _worker_loop(self, worker_id: str, streamer_socket) -> None:
        heatbeat_message = {_HEADER_KEY: _CTRL_MSG_HEADER, "topic_id": "ctl_msg/heartbeat", "data": {}, "timestamp": 0}
        fetch_message = {_HEADER_KEY: _CTRL_MSG_HEADER, "topic_id": "ctl_msg/fetch", "data": {}, "timestamp": 0}
        max_outgoing_message_queue_size = 0
        while self._run_worker_loop:
            # Broadcast any outgoing messages.
            while not self.outgoing_message_queue.empty():
                time_now = time.time()
                max_outgoing_message_queue_size = max(self.outgoing_message_queue.qsize(), max_outgoing_message_queue_size)
                message = self.outgoing_message_queue.get()
                queue_delay_time = time_now - message["timestamp"]
                self.stats_queue.put({
                    "outgoing_message_queue_latency": queue_delay_time,
                    "max_outgoing_message_queue_size": max_outgoing_message_queue_size
                })
                
                if message[_HEADER_KEY] == _CTRL_MSG_HEADER:
                    message_sent = self._send_control_message(message, streamer_socket)
                elif message[_HEADER_KEY] == _DATA_MSG_HEADER:
                    message_sent = self._send_data_message(message, streamer_socket)
                assert message_sent

            # Queue a heartbeat or fetch message if needed.
            time_now = time.time()
            if self._run_worker_loop and time_now - heatbeat_message["timestamp"] >= _HEARTBEAT_DELAY_SEC:
                heatbeat_message["timestamp"] = time_now
                self.outgoing_message_queue.put({_HEADER_KEY: _CTRL_MSG_HEADER, **heatbeat_message})
            if self._run_worker_loop and time_now - fetch_message["timestamp"] >= self.fetch_time_sec:
                fetch_message["timestamp"] = time_now
                self.outgoing_message_queue.put({_HEADER_KEY: _CTRL_MSG_HEADER, **fetch_message})

    def _handshake(self) -> bool:
        api_endpoint = self._get_endpoint(self.streamer_endpoint, self.streamer_channel, self.stream_uid)
        logging.info(f"Connecting to {api_endpoint}")
        for worker_id in range(self.num_workers):
            self.streamer_sockets.append(create_connection(api_endpoint))
        if self.post_connect_timeout_sec > 0:
            time.sleep(self.post_connect_timeout_sec)
        # Send and receive a control message to validate the connection.
        message = {_HEADER_KEY: _CTRL_MSG_HEADER, "topic_id": "ctl_msg/handshake", "data": {}, "timestamp": time.time()}
        for worker_id, streamer_socket in enumerate(self.streamer_sockets):
            if not self._send_control_message(message, streamer_socket):
                raise ValueError(f"Failed to handshake for socket {worker_id}")
        return True

    def _send_control_message(self, message: dict, streamer_socket) -> bool:
        assert message[_HEADER_KEY] == _CTRL_MSG_HEADER
        # Send the control message.
        assert self._send_data(message, streamer_socket) > 0
        # Get the control response.
        response_bytes = streamer_socket.recv()
        if not response_bytes:
            return False
        response = json.loads(response_bytes)
        if "topic_id" in response:
            if response["topic_id"].startswith("ctl_msg/"):
                logging.debug(f"Got control message: {response['topic_id']}")
            else:
                self.incoming_message_queue.put((response["topic_id"], response["data"]))
        elif "messages" in response:
            for message in response["messages"]:
                self.incoming_message_queue.put((message["topic_id"], message["message"]))
        elif "sensor_data" in response:
            for event in response["sensor_data"]:
                self.incoming_data_queue.put((event["sensor_name"], event["topic_id"], event["data"]))
        return True

    def _send_data_message(self, message: dict, streamer_socket) -> bool:
        """Sends a data message to the server, does not wait for a response."""
        assert message[_HEADER_KEY] == _DATA_MSG_HEADER
        start_time = time.time()
        num_bytes_sent = self._send_data(message, streamer_socket)
        assert num_bytes_sent > 0, f"Failed to send message!"
        end_time = time.time()
        latency = end_time - start_time
        topic_id = message["topic_id"]
        logging.debug(f"Sent topic_id: {topic_id} payload size: {num_bytes_sent} bytes latency: {latency}")
        self.outgoing_message_latency_total += latency
        self.outgoing_message_count += 1
        return True

    def _send_data(self, message: dict, streamer_socket) -> int:
        message_bytes = json.dumps(message).encode()
        streamer_socket.send_binary(message_bytes)
        num_bytes_sent = len(message_bytes)
        return num_bytes_sent


from typing import Any
import logging
import json
import time

from archetypeai._base import ApiBase
from archetypeai._socket_manager import SocketManager


class MessagingApi(ApiBase):
    """Main class for handling all messaging API calls."""

    def __init__(
        self,
        api_key: str,
        api_endpoint: str,
        client_name: str = "python_client",
        rate_limiter_timeout_sec: float = 0.5,
        fetch_time_sec=1.0) -> None:
        super().__init__(api_key, api_endpoint)
        self.client_name = client_name
        self.rate_limiter_timeout_sec = rate_limiter_timeout_sec
        self.fetch_time_sec = fetch_time_sec
        self.last_get_time = 0.0
        self.subscriber_info = []
        self.subscribers = []
    
    def subscribe(self, topic_ids: list[str]) -> dict:
        assert topic_ids, "Failed to subscribe, topic ids is empty!"
        api_endpoint = self._get_endpoint(self.api_endpoint, "messaging/subscribe")
        data_payload = {"client_name": self.client_name, "topic_ids": topic_ids}
        response = self.requests_post(api_endpoint, data_payload=json.dumps(data_payload))
        self.subscriber_info.append(response)

        new_subscriber = SocketManager(
            self.api_key, self.api_endpoint, num_worker_threads=1, fetch_time_sec=self.fetch_time_sec)
        new_subscriber._start_stream(response["subscriber_uid"], response["subscriber_endpoint"], "messaging")
        self.subscribers.append(new_subscriber)
        return response
    
    def close(self):
        """Closes and destroys any active subscribers."""
        for subscriber in self.subscribers:
            subscriber.close()
        self.subscribers = []

    def broadcast(self, topic_id: str, message: Any) -> dict:
        assert topic_id, "Failed to broadcast message, topic id is empty!"
        api_endpoint = self._get_endpoint(self.api_endpoint, "messaging/broadcast")
        data_payload = {"client_name": self.client_name, "messages": [{"topic_id": topic_id, "message": message}]}
        response = self.requests_post(api_endpoint, data_payload=json.dumps(data_payload))
        return response
    
    def get_next_messages(self) -> list[dict]:
        messages = []
        for subscriber in self.subscribers:
            for topic_id, message in subscriber.get_messages():
                messages.append({"topic_id": topic_id, "message": message})
        return messages


from typing import Dict, List, Tuple, Optional

import logging
from pathlib import Path
import secrets

import requests

from archetypeai._common import DEFAULT_ENDPOINT, safely_extract_response_data
from archetypeai._errors import ApiError


class ApiBase:
    """Base API functionality shared across all API modules."""

    def __init__(self,
                 api_key: str,
                 api_endpoint: str = DEFAULT_ENDPOINT,
                 num_retries: int = 3,
                 client_id: str = "",
                 request_timeout_sec: Optional[int] = None,
                 ) -> None:
        self.api_key = api_key
        self.api_endpoint = api_endpoint
        self.auth_headers = {"Authorization": f"Bearer {self.api_key}"}
        self.num_retries = num_retries
        self.valid_response_codes = (200, 201)
        self.invalid_response_codes = [error_code for error_code in range(400, 417)]
        self.client_id = client_id if client_id else secrets.token_hex(8)  # Generate a uid for this client.
        self.request_timeout_sec = request_timeout_sec
    
    def requests_get(self, api_endpoint: str, params: dict = {}, additional_headers: dict = {}) -> dict:
        request_args = {"api_endpoint": api_endpoint, "params": params, "additional_headers": additional_headers}
        return self._execute_request(request_func=self._requests_get, request_args=request_args)

    def _requests_get(self, api_endpoint: str, params: dict = {}, additional_headers: dict = {}) -> Tuple[int, dict]:
        response = requests.get(api_endpoint, params=params, headers={**self.auth_headers, **additional_headers})
        return response.status_code, safely_extract_response_data(response)
    
    def requests_post(self, api_endpoint: str, data_payload: bytes, additional_headers: dict = {}) -> dict:
        request_args = {"api_endpoint": api_endpoint, "data_payload": data_payload, "additional_headers": additional_headers}
        return self._execute_request(request_func=self._requests_post, request_args=request_args)

    def _requests_post(self, api_endpoint: str, data_payload: bytes, additional_headers: dict = {}) -> Tuple[int, dict]:
        response = requests.post(
            api_endpoint,
            data=data_payload,
            headers={**self.auth_headers, **additional_headers},
            timeout=self.request_timeout_sec)
        return response.status_code, safely_extract_response_data(response)

    def requests_delete(self, api_endpoint: str, params: dict = {}, additional_headers: dict = {}) -> dict:
        request_args = {"api_endpoint": api_endpoint, "params": params, "additional_headers": additional_headers}
        return self._execute_request(request_func=self._requests_delete, request_args=request_args)

    def _requests_delete(self, api_endpoint: str, params: dict = {}, additional_headers: dict = {}) -> Tuple[int, dict]:
        response = requests.delete(api_endpoint, params=params, headers={**self.auth_headers, **additional_headers})
        return response.status_code, safely_extract_response_data(response)

    def requests_download(self, api_endpoint: str, params: dict = {}, additional_headers: dict = {}) -> dict:
        request_args = {"api_endpoint": api_endpoint, "params": params, "additional_headers": additional_headers}
        return self._execute_request(request_func=self._requests_download, request_args=request_args)

    def _requests_download(self, api_endpoint: str, params: dict = {}, additional_headers: dict = {}) -> Tuple[int, requests.Response]:
        response = requests.get(
            api_endpoint,
            params=params,
            headers={**self.auth_headers, **additional_headers},
            timeout=self.request_timeout_sec)
        if response.status_code != 200:
            logging.warning(f"Failed to download file: {api_endpoint}. Error: {response}")
        return response.status_code, response

    def _execute_request(self, request_func, request_args: dict):
        num_attempts = 0
        while num_attempts < self.num_retries:
            response_code, response_data = request_func(**request_args)
            if response_code in self.valid_response_codes:
                return response_data
            if response_code in self.invalid_response_codes:
                raise ApiError(response_data)
            logging.warning(f"Failed to get valid response, got {response_code} {response_data} retrying...")
            num_attempts += 1
            if num_attempts >= self.num_retries:
                error_msg = f"Request failed after {num_attempts} attempts with error: {response_code} {response_data}"
                raise ValueError(error_msg)
    
    def _get_endpoint(self, base_endpoint: str, *args) -> str:
        subpath = None
        for arg in args:
            if subpath is None:
                subpath = arg
            else:
                subpath = Path(subpath) / Path(arg)
        protocol= ""
        potential_protocols = ("https://", "http://", "wss://", "ws://")
        for potential_protocol in potential_protocols:
            if base_endpoint.startswith(potential_protocol):
                protocol = potential_protocol
                break
        api_endpoint = protocol + str(Path(base_endpoint.replace(protocol, "")) / Path(subpath))
        api_endpoint = api_endpoint.replace("\\", "/")  # Needed for Windows style joins.
        return api_endpoint

    def get_file_type(self, filename: str) -> str:
        """Returns the file type of the input filename or raises an error."""
        file_ext = Path(filename).suffix.lower()
        file_ext_mapper = {
            '.png': 'image/png',
            '.jpg': 'image/jpeg',
            '.jpeg': 'image/jpeg',
            '.mp4': 'video/mp4',
            '.json': 'text/plain',
            '.jsonl': 'text/plain',
            '.csv': 'text/csv',
            '.txt': 'text/plain',
            '.text': 'text/plain',
        }
        if file_ext in file_ext_mapper:
            return file_ext_mapper[file_ext]
        raise ValueError(f"Unsupported file type: {file_ext}")

    def get_client_id(self) -> str:
        return self.client_id


from typing import Callable
import json
import logging
from queue import Queue
import threading
import time
import ast

import httpx
from httpx_sse import connect_sse


class ServerSideEventsReader:
    """Manages a threaded SSE reader."""

    def __init__(self, session_endpoint: str, header: dict, max_read_time_sec: float = -1.0, max_retries: int = 3):
        self.max_read_time_sec = max_read_time_sec
        self.max_retries = max_retries
        self.heartbeat_sec = 30
        self.continue_worker_loop = False
        self.read_event_queue = Queue()
        self.worker = threading.Thread(
            target=self._worker, args=(session_endpoint, header))
        self.worker.start()

    def __del__(self):
        self.close()

    def close(self) -> bool:
        """Stops and closes an active reader."""
        worker_stopped = False
        self.continue_worker_loop = False
        if self.worker is not None:
            self.continue_worker_loop = False
            self.worker.join()
            self.worker = None
            worker_stopped = True
        return worker_stopped

    def read(self, max_num_events: int = -1, block: bool = False):
        """Reads any queued events."""
        num_events_read = 0
        queue_not_empty = not self.read_event_queue.empty()
        keep_reading = True if block else queue_not_empty
        while keep_reading:
            if queue_not_empty or block:
                event = self.read_event_queue.get(block=block)
                yield event
                num_events_read += 1
            # Stop reading if we've reached the end of the events (in non-blocking mode)
            # or if we've reached the maximum number of events.
            queue_empty = not self.read_event_queue.empty()
            queue_not_empty = not queue_empty
            if queue_empty and not block:
                keep_reading = False
            if max_num_events > 0 and num_events_read >= max_num_events:
                keep_reading = False
            # Always stop the reader loop if the main worker loop has stopped.
            keep_reading &= self.continue_worker_loop
    
    def _worker(self, session_endpoint: str, header: dict) -> None:
        self.continue_worker_loop = True
        restart_delay_sec = 1
        num_retries = 0
        start_time = time.time()
        while self.continue_worker_loop:
            try:
                success = self._run_worker_loop(session_endpoint, header, start_time)
                if success:
                    num_retries = 0
            except Exception as exception:
                num_retries += 1
                if num_retries <= self.max_retries:
                    logging.exception("Failed to run reader loop - restarting...")
                    time.sleep(restart_delay_sec)
                    restart_delay_sec = max(restart_delay_sec * 2, 10)
                else:
                    logging.exception("Failed to run reader loop - reached max retries, stopping...")
                    self.continue_worker_loop = False
        self.continue_worker_loop = False
        self.worker = None

    def _run_worker_loop(self, session_endpoint: str, header: dict, start_time: float) -> bool:
        """Connects to and reads events from an SSE remote connection until instructed to stop."""
        logging.info(f"[sse reader] Connecting to {session_endpoint}")
        headers = {**header, "Accept": "text/event-stream"}
        num_events_read = 0
        with httpx.Client() as client:
            with connect_sse(client, "GET", session_endpoint, headers=headers, timeout=10.0) as event_source:
                # Try and read any SSE events, this will block until an event is received.
                for event in event_source.iter_sse():
                    assert event.event == "message", event
                    try:
                        logging.debug(event)
                        event_data = json.loads(event.data)

                        assert "type" in event_data, event
                        self.read_event_queue.put(event_data)
                        num_events_read += 1
                        if event_data["type"] == "sse.stream.heartbeat":
                            continue
                        if event_data["type"] == "sse.stream.end":
                            # Cancel the worker loop so the thread will gracefully stop.
                            self.continue_worker_loop = False
                            break
                        if self.max_read_time_sec >= 0 and time.time() - start_time >= self.max_read_time_sec:
                            self.continue_worker_loop = False
                            break
                    except Exception as exception:
                        logging.debug(f"Failed to parse JSON packet: {event}")

                    if not self.continue_worker_loop:
                        logging.info(f"[sse reader] Received stop signal...")
                        break

        current_time = time.time()
        run_time = current_time - start_time
        logging.info(f"[sse reader] Reached end of stream. num_events: {num_events_read} run_time: {run_time:.2f} sec")

        return True


from typing import Dict, Union

import logging
import os
import json

from requests_toolbelt import MultipartEncoder

from archetypeai._base import ApiBase


class FilesApiBase(ApiBase):
    """Common file ops shared across all file APIs."""

    def get_info(self) -> dict:
        """Gets the file info for all the files your org has uploaded to the Archetype AI platform."""
        api_endpoint = self._get_endpoint(self.api_endpoint, "files/info")
        return self.requests_get(api_endpoint)

    def get_metadata(self, shard_index: int = -1, max_items_per_shard: int = -1, file_id: Union[str, None] = None) -> dict:
        """Gets a list of metadata about any files your org has uploaded to the Archetype AI platform.

        Use the shard_index and max_items_per_shard to retrieve metadata about a subset of files.

        Use the file_id argument to retrieve the metadata for a specific file.
        """
        if file_id is None:
            api_endpoint = self._get_endpoint(self.api_endpoint, "files/metadata")
            params = {"shard_index": shard_index, "max_items_per_shard": max_items_per_shard}
        else:
            api_endpoint = self._get_endpoint(self.api_endpoint, f"files/metadata/{file_id}")
            params = {}
        return self.requests_get(api_endpoint, params=params)


class LocalFilesApi(FilesApiBase):
    """API for working with local files."""

    def upload(self, filename: str, base64_data: Union[str, None] = None) -> dict:
        """Uploads a local file to the Archetype AI platform."""
        if base64_data is None:
            response = self._upload_file(filename)
        else:
            response = self._upload_base64_data(filename, base64_data)
        assert "file_id" in response, response
        return response
        
    def _upload_file(self, filename: str) -> dict:
        api_endpoint = self._get_endpoint(self.api_endpoint, "files")
        with open(filename, "rb") as file_handle:
            encoder = MultipartEncoder(
                {"file": (os.path.basename(filename), file_handle.read(), self.get_file_type(filename))})
            response_data = self.requests_post(
                api_endpoint, data_payload=encoder, additional_headers={"Content-Type": encoder.content_type}
            )
            return response_data
    
    def _upload_base64_data(self, filename: str, base64_data: str) -> dict:
        api_endpoint = self._get_endpoint(self.api_endpoint, "files/base64")
        encoder = MultipartEncoder(
            {"file": (os.path.basename(filename), base64_data, self.get_file_type(filename))})
        response_data = self.requests_post(
            api_endpoint, data_payload=encoder, additional_headers={"Content-Type": encoder.content_type}
        )
        return response_data

    def delete(self, filename: str) -> dict:
        """Deletes a file that was previously uploaded to the Archetype AI platform."""
        api_endpoint = self._get_endpoint(self.api_endpoint, f"files/delete/{filename}")
        return self.requests_delete(api_endpoint)

    def download(self, filename: str, local_filename: str = "") -> bool:
        """Downloads a file that was previously uploaded to the Archetype AI platform."""
        api_endpoint = self._get_endpoint(self.api_endpoint, f"files/download/{filename}")
        # If the local filename is not set then download the file using the remote filename.
        if local_filename == "":
            local_filename = filename
        response = self.requests_download(api_endpoint)
        success = response.status_code == 200
        if success:
            with open(local_filename, "wb") as file_handle:
                    file_handle.write(response.content)
        return success


class S3FilesApi(FilesApiBase):
    """API for working with Amazon s3 files."""

    def upload(self, filenames: list[str], credentials: Union[Dict[str, str], None] = None, batch_size: int = 64) -> dict:
        """Uploads a list of s3 files directly to the Archetype AI platform."""
        api_endpoint = self._get_endpoint(self.api_endpoint, "files/s3")
        num_files = len(filenames)
        response_data = []
        for batch_start in range(0, num_files, batch_size):
            batch_end = min(batch_start + batch_size, num_files)
            data_payload = {"credentials": credentials, "filenames": filenames[batch_start:batch_end]}
            response_data.append(self.requests_post(api_endpoint, data_payload=json.dumps(data_payload)))
        return response_data


class FilesApi(FilesApiBase):
    """Main class for handling all file API calls."""

    local: LocalFilesApi
    s3: S3FilesApi

    def __init__(self, api_key: str, api_endpoint: str) -> None:
        super().__init__(api_key, api_endpoint)
        self.local = LocalFilesApi(api_key, api_endpoint)
        self.s3 = S3FilesApi(api_key, api_endpoint)


from typing import Dict
import inspect

import requests

DEFAULT_ENDPOINT = "https://api.u1.archetypeai.app/v0.5"


def safely_extract_response_data(response: requests.Response) -> Dict:
    """Safely extracts the response data from both valid and invalid responses."""
    try:
        response_data = response.json()
        return response_data
    except:
        return {}


def filter_kwargs(func, kwarg_dict):
    """Filters kwargs based on the signature of the input function."""
    sign = set([val.name for val in inspect.signature(func).parameters.values()])
    filtered_dict = {key: kwarg_dict[key] for key in sign.intersection(kwarg_dict.keys())}
    return filtered_dict


import json

from kafka import KafkaConsumer
from kafka import KafkaProducer

from archetypeai._base import ApiBase


class KafkaMessageConsumer:
    """Wrapper class for consuming kafka messages from a secure topic_uid."""

    def __init__(
        self,
        kafka_broker_endpoints: list[str],
        topic_uid: str,
        auto_offset_reset: str,
        consumer_timeout_ms: int,
        consumer_message_batch_size: int,
    ) -> None:
        self.consumer = KafkaConsumer(
            topic_uid,
            bootstrap_servers=kafka_broker_endpoints,
            auto_offset_reset=auto_offset_reset,
            consumer_timeout_ms=consumer_timeout_ms,
            value_deserializer=lambda x: json.loads(x.decode('utf-8'))
        )
        self.consumer_message_batch_size = consumer_message_batch_size

    def __iter__(self):
        batch_size = self.consumer_message_batch_size
        for batch_index, message in enumerate(self.consumer):
            yield message
            if batch_size > 0 and batch_index >= batch_size:
                break


class KafkaMessageProducer:
    """Wrapper class for producing kafka messages under a secure topic_uid."""

    def __init__(self, kafka_broker_endpoints: list[str], topic_id_map: dict) -> None:
        self.producer = KafkaProducer(
            bootstrap_servers=kafka_broker_endpoints,
            value_serializer=lambda x: json.dumps(x).encode('utf-8')
        )
        self.topic_id_map = topic_id_map

    def send(self, topic_id: str, value: dict) -> bool:
        assert topic_id in self.topic_id_map
        topic_uid = self.topic_id_map[topic_id]
        self.producer.send(topic_uid, value=value)
        return True

    def flush(self) -> bool:
        self.producer.flush()
        return True


class KafkaApi(ApiBase):
    """Main class for handling all kafka API calls."""

    def __init__(self, api_key: str, api_endpoint: str) -> None:
        super().__init__(api_key, api_endpoint)
    
    def create_topic(self, topic_id: str) -> dict:
        """Creates a new topic on the Archetype AI kafka service."""
        api_endpoint = self._get_endpoint(self.api_endpoint, "kafka/topics/create")
        data_payload = {"topic_id": topic_id}
        response = self.requests_post(api_endpoint, data_payload=json.dumps(data_payload))
        return response

    def subscribe_topic(self, topic_id: str) -> dict:
        """Subscribes to an existing topic on the Archetype AI kafka service."""
        api_endpoint = self._get_endpoint(self.api_endpoint, "kafka/topics/subscribe")
        data_payload = {"topic_id": topic_id}
        response = self.requests_post(api_endpoint, data_payload=json.dumps(data_payload))
        return response

    def create_producer(self, topic_ids: list[str]) -> KafkaMessageProducer:
        """Creates the topic_ids on the Archetype kafka service and returns a producer for them."""
        assert topic_ids, "Empty topic ids!"
        kafka_broker_endpoints = set()
        topic_id_map = {}
        for topic_id in topic_ids:
            response = self.create_topic(topic_id)
            topic_id_map[topic_id] = response["topic_uid"]
            for kafka_broker_endpoint in response["kafka_broker_endpoints"]:
                kafka_broker_endpoints.add(kafka_broker_endpoint)
        kafka_broker_endpoints = list(kafka_broker_endpoints)
        producer = KafkaMessageProducer(kafka_broker_endpoints, topic_id_map)
        return producer

    def create_consumer(
        self,
        topic_id: str,
        auto_offset_reset: str = "earliest",
        consumer_timeout_ms: int = 30000,
        consumer_message_batch_size: int = 16,
        ) -> KafkaMessageConsumer:
        """Subscribes to the topic_id on the Archetype kafka service and returns a consumer for them."""
        assert topic_id, "Empty topic id!"
        response = self.subscribe_topic(topic_id)
        topic_uid = response["topic_uid"]
        kafka_broker_endpoints = response["kafka_broker_endpoints"]
        consumer = KafkaMessageConsumer(
            kafka_broker_endpoints,
            topic_uid,
            auto_offset_reset=auto_offset_reset,
            consumer_timeout_ms=consumer_timeout_ms,
            consumer_message_batch_size=consumer_message_batch_size,
        )
        return consumer


import logging

from archetypeai._base import ApiBase
from archetypeai._capabilities import CapabilitiesApi
from archetypeai._common import DEFAULT_ENDPOINT, filter_kwargs
from archetypeai._errors import ApiError as ApiError
from archetypeai._files import FilesApi
from archetypeai._messaging import MessagingApi
from archetypeai._sensors import SensorsApi
from archetypeai._lens import LensApi
from archetypeai._kafka_client import KafkaApi

_VERSION = "26.02.24.0"


class ArchetypeAI(ApiBase):
    """Main client for the Archetype AI platform."""

    files: FilesApi
    capabilities: CapabilitiesApi
    messaging: MessagingApi
    sensors: SensorsApi
    lens: LensApi
    kafka: KafkaApi

    @staticmethod
    def get_version() -> str:
        """Returns the current version of the Archetype client."""
        return _VERSION
    
    @staticmethod
    def get_default_endpoint() -> str:
        """Returns the default endpoint the Archetype client should connect to."""
        return DEFAULT_ENDPOINT

    def __init__(self, api_key: str, api_endpoint: str = DEFAULT_ENDPOINT, **kwargs) -> None:
        super().__init__(api_key, api_endpoint)
        input_args = {"api_key": api_key, "api_endpoint": api_endpoint, **kwargs}
        self.files = FilesApi(**filter_kwargs(FilesApi.__init__, input_args))
        self.capabilities = CapabilitiesApi(**filter_kwargs(CapabilitiesApi.__init__, input_args))
        self.messaging = MessagingApi(**filter_kwargs(MessagingApi.__init__, input_args))
        self.sensors = SensorsApi(**filter_kwargs(SensorsApi.__init__, input_args))
        self.lens = LensApi(**filter_kwargs(LensApi.__init__, input_args))
        self.kafka = KafkaApi(**filter_kwargs(KafkaApi.__init__, input_args))


import json

from archetypeai._base import ApiBase


class DataProcessingApi(ApiBase):
    """Main class for handling all data processing API calls."""

    def __init__(self, api_key: str, api_endpoint: str) -> None:
        super().__init__(api_key, api_endpoint)

    def create_job(self, job_config: dict) -> dict:
        """Creates a new data processing job."""
        api_endpoint = self._get_endpoint(self.api_endpoint, "data_processing")
        data_payload = {"job_config": job_config}
        response_data = self.requests_post(api_endpoint, data_payload=json.dumps(data_payload))
        return response_data
    
    def get_info(self) -> dict:
        """Returns the high-level information about all data processing jobs in your organization."""
        api_endpoint = self._get_endpoint(self.api_endpoint, "data_processing/info")
        response_data = self.requests_get(api_endpoint)
        return response_data

    def get_metadata(self, shard_index: int = -1, max_items_per_shard: int = -1) -> dict:
        """Gets a list of metadata about any data processing jobs in your organization.

        Use the shard_index and max_items_per_shard to retrieve information about a subset of jobs.
        """
        api_endpoint = self._get_endpoint(self.api_endpoint, "data_processing/metadata")
        params = {"shard_index": shard_index, "max_items_per_shard": max_items_per_shard}
        return self.requests_get(api_endpoint, params=params)



class ApiError(Exception):
    def __init__(self, api_error):
        super().__init__(api_error)
        if isinstance(api_error, dict) and "errors" in api_error:
            self.errors = api_error["errors"]
        else:
            self.errors = None
    
    def validate_error_code_exists(self, error_code: str) -> bool:
        """Validates that at least one instance of the error code exists."""
        if self.errors is None:
            return False
        for error in self.errors:
            code = error.get("code", None)
            if code == error_code:
                return True
        return False
    
    def validate_error_count(self, error_count: int) -> bool:
        """Validates the number of expected error messages match the count."""
        if self.errors is None:
            return False
        return len(self.errors) == error_count


from typing import Optional

import argparse
import base64
import logging
import yaml
import sys

from archetypeai._common import DEFAULT_ENDPOINT


def base64_encode(filename: str) -> str:
    with open(filename, "rb") as img_handle:
        encoded_img = base64.b64encode(img_handle.read()).decode("utf-8")
    return encoded_img


def pformat(data: dict, prefix: str = "") -> str:
    """Prints a dictonary as a formatted yaml string."""
    yaml_string = yaml.dump(data, sort_keys=False, default_flow_style=False)
    fomatted_string = f"{prefix}{yaml_string}"
    return fomatted_string


class ArgParser(argparse.ArgumentParser):
    """Creates a custom arg parser with common ArchetypeAI args."""

    def __init__(self):
        super().__init__(self)
        self.add_argument("--api_key", required=True, type=str, help="Your Archetype AI API key")
        self.add_argument("--api_endpoint", default=DEFAULT_ENDPOINT, type=str, help="The target API endpoint")
        self.add_argument("--logging_level", default=logging.INFO, type=str, help="Sets the logging level")

    def parse_args(self, configure_logging: Optional[bool] = None) -> argparse.Namespace:
        """Configures default logging using optional args."""
        # Call the default arg parser.
        args = super().parse_args()

        # Add a basic logging config if enable.
        if configure_logging:
            logging.basicConfig(level=args.logging_level, format="[%(asctime)s] %(message)s", datefmt="%H:%M:%S", stream=sys.stdout)
        
        return args


import json
import logging
from queue import Queue
import threading
import time

import websocket


class LensSessionSocket:
    """Manages websocket connections for each lens session."""

    def __init__(self, session_endpoint: str, header: dict):
        self.heartbeat_sec = 30
        self.max_worker_restarts = 10
        self.run_worker = False
        self.read_event_queue = Queue()
        self.write_event_queue = Queue()
        self.worker = threading.Thread(
            target=self._worker, args=(session_endpoint, header))
        self.worker.start()

    def __del__(self):
        self.close()

    def close(self) -> bool:
        """Stops and closes an active socket."""
        worker_stopped = False
        if self.run_worker:
            self.run_worker = False
            self.worker.join()
            self.worker = None
            worker_stopped = True
        return worker_stopped

    def send_and_recv(self, event_data: dict) -> dict:
        """Writes an event to an open session and returns the response."""
        self.write_event_queue.put(event_data)
        response = self.read_event_queue.get()
        response = json.loads(response)
        return response
    
    def _worker(self, session_endpoint: str, header: dict):
        self.run_worker = True
        num_restarts = 0
        if "User-Agent" not in header:
            header["User-Agent"] = "archetypeai.py"
        while self.run_worker:
            try:
                self.run_worker = self._run_worker_loop(session_endpoint, header)
            except Exception as exception:
                num_restarts += 1
                if num_restarts < self.max_worker_restarts:
                    logging.exception("Failed to run socket loop - restarting...")
                else:
                    logging.exception("Failed to run socket loop...")
                    self.run_worker = False
        self.run_worker = False

    def _run_worker_loop(self, session_endpoint: str, header: dict) -> bool:
        socket = websocket.create_connection(session_endpoint, header=header)
        start_time = time.time()
        last_event = time.time()
        while self.run_worker:
            current_time = time.time()
            run_time = current_time - start_time
            if not self.write_event_queue.empty():
                last_event = time.time()
                # Send the event to the server.
                event_data = self.write_event_queue.get()
                logging.debug(f"[{run_time:.2f}] Sending event w/ type: {event_data['type']}...")
                socket.send_binary(json.dumps(event_data).encode())
                # Read back the response.
                event_data = socket.recv()
                if event_data:
                    self.read_event_queue.put(event_data)
                else:
                    logging.warning(f"Received empty event: {event_data}")
            else:
                time.sleep(0.1)
            # Send a periodic heartbeat to keep the connection alive.
            if current_time - last_event >= self.heartbeat_sec:
                self.write_event_queue.put({"type": "session.heartbeat"})
        return self.run_worker


import re
import ast
from setuptools import setup, find_packages

# Dynamically extract the library version from the client.
_version_re = re.compile(r'_VERSION\s*=\s*(.*)')
with open("src/archetypeai/api_client.py", 'rb') as f:
    match = _version_re.search(f.read().decode('utf-8'))
    if match:
        version = str(ast.literal_eval(match.group(1)))
    else:
        raise RuntimeError("Unable to find version string.")

setup(
    name="archetypeai",
    version=version,
    author="Archetype AI",
    url="https://github.com/archetypeai/python-client",
    description="The official python client for the Archetype AI API.",
    packages=find_packages(where="src"),
    package_dir={"": "src"},
    python_requires=">=3.8",
    install_requires=[
        "wheel",
        "argparse>=1.4.0",
        "typing-extensions>=4.8.0",
        "requests>=2.31.0",
        "requests-toolbelt>=1.0.0",
        "websockets>=12.0",
        "websocket-client>=1.8.0",
        "kafka-python==2.0.4",
        "httpx==0.28.1",
        "httpx-sse==0.4.1",
        "pyyaml==6.0.2",
        "pytest==9.0.1",
    ],
    include_package_data=True,
)



from .api_client import ArchetypeAI
from .utils import ArgParser, pformat
from ._errors import ApiError

__all__ = ["ArchetypeAI", "ApiError", "ArgParser", "pformat"]
__version__ = ArchetypeAI.get_version()



from typing import Callable
import json
import logging
import time

import yaml

from archetypeai._base import ApiBase
from archetypeai._lens_session_socket import LensSessionSocket
from archetypeai._sse import ServerSideEventsReader


class SessionsApi(ApiBase):
    """Main class for handling all lens session API calls."""

    session_socket_cache: dict = {}

    def __init__(self, api_key: str, api_endpoint: str) -> None:
        super().__init__(api_key, api_endpoint)

    def __del__(self):
        self.close()

    def get_info(self) -> dict:
        """Gets the high-level info for all lens sessions across your org."""
        api_endpoint = self._get_endpoint(self.api_endpoint, "lens/sessions/info")
        return self.requests_get(api_endpoint)

    def get_metadata(self, shard_index: int = -1, max_items_per_shard: int = -1, session_id: str = "") -> dict:
        """Gets a list of metadata about any lens sessions across your org.

        Use the shard_index and max_items_per_shard to retrieve information about a subset of sessions.

        To request metadata about a specific session, pass just the session_id.
        """
        api_endpoint = self._get_endpoint(self.api_endpoint, "lens/sessions/metadata")
        params = {"shard_index": shard_index, "max_items_per_shard": max_items_per_shard, "session_id": session_id}
        return self.requests_get(api_endpoint, params=params)

    def create(self, lens_id: str) -> dict:
        api_endpoint = self._get_endpoint(self.api_endpoint, "lens/sessions/create")
        data = {"lens_id": lens_id}
        response = self.requests_post(api_endpoint, data_payload=json.dumps(data))
        return response

    def destroy(self, session_id: str) -> dict:
        assert session_id, "Failed to destroy, session_id is empty!"
        api_endpoint = self._get_endpoint(self.api_endpoint, "lens/sessions/destroy")
        data = {"session_id": session_id}
        response = self.requests_post(api_endpoint, data_payload=json.dumps(data))
        return response

    def connect(self, session_id: str, session_endpoint: str) -> bool:
        try:
            socket = LensSessionSocket(session_endpoint, {"Authorization":f"Bearer {self.api_key}"})
            self.session_socket_cache[session_id] = socket
        except Exception as exception:
            logging.exception(f"Failed to connect to session at {session_endpoint}")
            return False
        return True
    
    def read(self, session_id: str, client_id: str = "") -> list[dict]:
        """Reads an event from an open session and returns the response."""
        assert session_id in self.session_socket_cache, f"Unknown session ID {session_id}"
        client_id = client_id if client_id else self.client_id
        event_data = {"type": "session.read", "event_data": {"client_id": client_id}}
        response = self.write(session_id, event_data)
        return response
    
    def write(self, session_id: str, event_data: dict) -> dict:
        """Writes an event to an open session and returns the response."""
        assert session_id in self.session_socket_cache, f"Unknown session ID {session_id}"
        response = self.session_socket_cache[session_id].send_and_recv(event_data)
        return response

    def process_event(self, session_id: str, event: dict) -> dict:
        api_endpoint = self._get_endpoint(self.api_endpoint, "lens/sessions/events/process")
        data = {"session_id": session_id, "event": event}
        response = self.requests_post(api_endpoint, data_payload=json.dumps(data))
        return response

    def create_sse_consumer(self, session_id: str, max_read_time_sec: float = -1.0) -> ServerSideEventsReader:
        """Creates a new server-side-event consumer and starts it in a background thread."""
        api_endpoint = self._get_endpoint(self.api_endpoint, f"lens/sessions/consumer/{session_id}")
        headers = {"Authorization":f"Bearer {self.api_key}"}
        sse_consumer = ServerSideEventsReader(api_endpoint, headers, max_read_time_sec)
        return sse_consumer
    
    def close(self) -> bool:
        """Closes and removes any open session socket. Returns true if any sessions were closed, false otherwise."""
        sessions_closed = False
        if self.session_socket_cache:
            for session_id in self.session_socket_cache:
                self.session_socket_cache[session_id].close()
            self.session_socket_cache = {}
            sessions_closed = True
        return sessions_closed 


class LensApi(ApiBase):
    """Main class for handling all lens API calls."""

    sessions: SessionsApi

    def __init__(self, api_key: str, api_endpoint: str) -> None:
        super().__init__(api_key, api_endpoint)
        self.sessions = SessionsApi(api_key, api_endpoint)

    def get_info(self) -> dict:
        """Gets the high-level info for all lenses across your org."""
        api_endpoint = self._get_endpoint(self.api_endpoint, "lens/info")
        return self.requests_get(api_endpoint)

    def get_metadata(self, shard_index: int = -1, max_items_per_shard: int = -1, lens_id: str = "") -> dict:
        """Gets a list of metadata about any lenses across your org.

        Use the shard_index and max_items_per_shard to retrieve information about a subset of lenses.

        To request metadata about a specific lens, pass just the lens_id.
        """
        api_endpoint = self._get_endpoint(self.api_endpoint, "lens/metadata")
        params = {"shard_index": shard_index, "max_items_per_shard": max_items_per_shard, "lens_id": lens_id}
        return self.requests_get(api_endpoint, params=params)

    def register(self, lens_config: dict) -> dict:
        """Registers a new lens with the Archetype AI platform."""
        api_endpoint = self._get_endpoint(self.api_endpoint, "lens/register")
        data = {"lens_config": lens_config}
        response = self.requests_post(api_endpoint, data_payload=json.dumps(data))
        lens_id = response.get("lens_id", None)
        assert lens_id, f"Missing lens_id: {response}"
        return response

    def clone(self, lens_id: str) -> dict:
        """Clones an existing lens to create a new custom lens."""
        api_endpoint = self._get_endpoint(self.api_endpoint, "lens/clone")
        data = {"lens_id": lens_id}
        response = self.requests_post(api_endpoint, data_payload=json.dumps(data))
        lens_id = response.get("lens_id", None)
        assert lens_id, f"Missing lens_id: {response}"
        return response

    def modify(self, lens_id: str, lens_params: dict) -> dict:
        """Modifies an existing lens and returns the results."""
        api_endpoint = self._get_endpoint(self.api_endpoint, "lens/modify")
        data = {"lens_id": lens_id, **lens_params}
        response = self.requests_post(api_endpoint, data_payload=json.dumps(data))
        return response

    def delete(self, lens_id: str) -> dict:
        """Deletes an existing lens and returns the results."""
        api_endpoint = self._get_endpoint(self.api_endpoint, "lens/delete")
        data = {"lens_id": lens_id}
        response = self.requests_post(api_endpoint, data_payload=json.dumps(data))
        return response

    def create_and_run_lens(
        self,
        lens_config: str | dict,
        session_fn: Callable,
        auto_destroy_lens: bool = True,
        auto_destroy_session: bool = True,
        **session_kwargs
        ):
        """Creates a new lens and automatically launches a new lens session."""

        # If the lens is passed as a str then dynamically convert it to a dict.
        if isinstance(lens_config, str):
            lens_config = yaml.safe_load(lens_config)
        assert isinstance(lens_config, dict), f"Invalid input: {lens_config}"

        # Register the custom lens with the Archetype AI platform.
        lens_metadata = self.register(lens_config)
        assert "lens_id" in lens_metadata, f"Missing lens id: {lens_metadata}"
        lens_id = lens_metadata["lens_id"]

        fn_response = self.create_and_run_session(
            lens_id, session_fn, auto_destroy=auto_destroy_session, **session_kwargs)

        if auto_destroy_lens:
            # Delete the custom lens to clean things up.
            self.delete(lens_id)

        return fn_response

    def create_and_run_session(
        self,
        lens_id: str,
        session_fn: Callable,
        auto_destroy: bool = True,
        **session_kwargs
        ):
        """Creates and runs a lens session based on a pre-existing lens."""
        # Create a new session based on this lens.
        session_id, session_endpoint = self.create_session(lens_id)

        fn_response = None
        try:
            # Connect to the lens and run the custom session.
            fn_response = session_fn(session_id, session_endpoint, **session_kwargs)
        finally:
            if auto_destroy:
                # Clean up the lens at the end of the session.
                self.destroy_lens_session(session_id)

        return fn_response

    def create_session(self, lens_id: str):
        """Creates a session and returns the session_id and session_endpoint."""
        try:
            logging.debug(f"Creating lens with id: {lens_id}")
            response = self.sessions.create(lens_id)
            logging.debug(f"{response}")
            if "errors" in response:
                raise ValueError(response["errors"])
        except ValueError as err:
            logging.exception(f"{err}")
            raise

        # Extract the session_id and endpoint.
        session_id = response["session_id"]
        session_endpoint = response["session_endpoint"]
        logging.info(f"Session ID: {session_id} @ {session_endpoint}")
        return session_id, session_endpoint

    def destroy_lens_session(self, session_id: str) -> None:
        """Cleanly stops and destroys an active session."""
        try:
            # Destroy any active sessions.
            response = self.sessions.destroy(session_id)
            logging.debug(response)
            logging.info(f"Session Status: {response['session_status']}")
        except:
            logging.exception("Failed to destroy sessions!")
        # Close any open sessions sockets.
        self.sessions.close()
